<?php

namespace Drupal\queue_scheduler\Plugin\QueueUI;

use Drupal\queue_ui\Plugin\QueueUI\DatabaseQueue;
use Drupal\Core\Database\Connection;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Defines the default Drupal Queue UI backend.
 *
 * @QueueUI(
 *   id = "database_queue_scheduler",
 *   class_name = "DatabaseQueueScheduler"
 * )
 */
class DatabaseQueueScheduler extends DatabaseQueue {

  /**
   * {@inheritdoc}
   *
   * @param \Symfony\Component\DependencyInjection\ContainerInterface $container
   *   The current service container.
   * @param array $configuration
   *   The configuration to use.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   *
   * @return \Drupal\Core\Plugin\ContainerFactoryPluginInterface|\Drupal\queue_ui\Plugin\QueueUI\DatabaseQueue
   *   Defines the default Drupal Queue UI backend
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('database')
    );
  }

  /**
   * DatabaseQueue constructor.
   *
   * @param array $configuration
   *   The configuration to use.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Database\Connection $database
   *   The database connection.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, Connection $database) {
    $this->database = $database;
  }

  /**
   * Inspect the queue items in a specified queue.
   *
   * @param string $queueName
   *   The name of the queue being inspected.
   *
   * @return mixed
   *   Return call the execute method.
   */
  public function getItems($queueName) {
    $query = $this->database->select('queue', 'q');
    $query->addField('q', 'item_id');
    $query->addField('q', 'expire');
    $query->addField('q', 'created');
    $query->addField('q', 'process_at');
    $query->condition('q.name', $queueName);
    $query = $query->extend('Drupal\Core\Database\Query\PagerSelectExtender');
    $query = $query->limit(25);

    return $query->execute();
  }

  /**
   * Load a specified SystemQueue queue item from the database.
   *
   * @param int $item_id
   *   The item id to load.
   *
   * @return mixed
   *   Result of the database query loading the queue item.
   */
  public function loadItem($item_id) {
    // Load the specified queue item from the queue table.
    $query = $this->database->select('queue', 'q')
      ->fields('q', [
        'item_id',
        'name',
        'data',
        'expire',
        'created',
        'process_at',
      ])
      ->condition('q.item_id', $item_id)
      // Item id should be unique.
      ->range(0, 1);

    return $query->execute()->fetchObject();
  }

  public function releaseItems($queueName) {
    return $this->database->update('queue')
      ->fields([
        'expire' => 0,
      ])
      ->condition('name', $queueName, '=')
      ->execute();
  }

  /**
   * Force the releasing of a specified queue item.
   *
   * @param int $item_id
   *   The item id to be released.
   */
  public function releaseItem($item_id) {
    $this->database->update('queue')
      ->condition('item_id', $item_id)
      ->fields(['expire' => 0])
      ->execute();
  }

  /**
   * Force the deletion of a specified queue item.
   *
   * @param int $item_id
   *   The item id to be deleted.
   */
  public function deleteItem($item_id) {
    $this->database->delete('queue')
      ->condition('item_id', $item_id)
      ->execute();
  }

}
